﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace PV178.Tutorial3.SpringfieldUniversityFull
{
    public class University
    {
        private const char FileValuesSeparator = ',';

        private IList<Student> students = new List<Student>();
        private IList<Seminar> seminars = new List<Seminar>();

        public IList<Student> Students { get { return students; } }
        public IList<Seminar> Seminars { get { return seminars; } }


        public University(params Seminar[] seminars)
        {
            foreach (var s in seminars)
            {
                this.seminars.Add(s);
            };
        }

        /*
         * Generator
         * 
         * Pomoci klicoveho slova yield lze vytvorit metodu, ktera se chova jako 
         * iterator. Kompilator sam vygeneruje prislusnou tridu pro popsany iterator.
         * 
         * yield return <vyraz>;   Oznacuje vraceni hodnoty z iteratoru a zaroven
         *                         si zapamatuje pozici v programu iteratoru, kde
         *                         skoncil. V dalsi iteraci pokracuje od tohoto
         *                         mista dale.
         *                           
         * yield break;            Slouzi pro okamzite ukonceni iterace. Na rozdil 
         *                         samotneho break se okamzite ukonci cely iterator,
         *                         ne jen prislusny cyklus.
         * 
         */

        public IEnumerable<Student> GetBestStudents()
        {
            foreach (var student in this.students.Where(s => s.Marks.Any(x => x.Value.HasValue)))
            {
                if (student.Marks.Where(m => m.Value.HasValue).Average(x => x.Value.ToScore()) < 1.2)
                {
                    yield return student;
                }
            }
        }

        public IEnumerable<Student> GetWorstStudents()
        {
            int count = 0;
            foreach (var student in this.students.Where(s => s.Marks.Any(x => x.Value.HasValue)))
            {
                if (student.Marks.Where(m => m.Value.HasValue).Average(x => x.Value.ToScore()) > 3)
                {
                    yield return student;
                    count++;
                }
            }
        }

        /*
         * LINQ
         */

        public IEnumerable<Student> StudentsWithMissingMarks()
        {
            return from s in students
                   where s.Marks.Any(x => x.Value.HasValue == false)
                   select s;
        }

        public void GetStatistics()
        {
            var query = students
                .GroupBy(s => s.LastName[0])
                .OrderBy(g => g.Key)
                .Select(g => new
                {
                    FirstCharacterOfLastName = g.Key,
                    Students = g
                });

            foreach (var item in query)
            {
                Console.WriteLine(item.FirstCharacterOfLastName);
                foreach (var student in item.Students)
                {
                    Console.WriteLine("- {0}, {1} ({2})", student.LastName, student.FirstName, student.Uco);
                }
            }
        }

        public void GetStatistics2()
        {
            var query = from s in students
                        from m in s.Marks
                        where m.Value.HasValue
                        group m by m.Key into marksBySeminar
                        select new { Seminar = marksBySeminar.Key, Marks = marksBySeminar };

            foreach (var item in query)
            {
                Console.WriteLine(item.Seminar);
                Console.WriteLine("- Average: {0}", item.Marks.Average(m => m.Value.ToScore()));
                Console.WriteLine("- Best:    {0}", item.Marks.Min(m => m.Value));
                Console.WriteLine("- Worst:   {0}", item.Marks.Max(m => m.Value));
            }
        }

        /*
         * File Operations
         *
         */

        public void LoadSeminars(string filename)
        {
            FileStream dataFile = null;
            StreamReader reader = null;
            try
            {
                dataFile = new FileStream(filename, FileMode.Open, FileAccess.Read);
                reader = new StreamReader(dataFile);

                string line = reader.ReadLine();
                while (line != null)
                {
                    string[] csv = line.Split(new char[] { FileValuesSeparator }, 2);
                    this.seminars.Add(new Seminar(csv[0], csv[1]));
                    line = reader.ReadLine();
                }
            }
            finally
            {
                if (reader != null) reader.Close();
                if (dataFile != null) dataFile.Close();
            }
        }

        public void SaveSeminars(string filename)
        {
            StreamWriter writer = new StreamWriter(filename);

            foreach (var item in seminars)
            {
                writer.WriteLine(item.Code + FileValuesSeparator + item.Name);
            }

            writer.Close();
        }

        public void LoadFromFile(string filename)
        {
            FileStream dataFile = null;
            StreamReader reader = null;
            try
            {
                dataFile = new FileStream(filename, FileMode.Open, FileAccess.Read);
                reader = new StreamReader(dataFile);

                string line = reader.ReadLine();
                while (line != null)
                {
                    ParseLine(line);
                    line = reader.ReadLine();
                }
            }
            finally
            {
                if (reader != null) reader.Close();
                if (dataFile != null) dataFile.Close();
            }
        }

        private void ParseLine(string line)
        {
            string[] lineParts = line.Split(FileValuesSeparator);

            Student newStudent = new Student
            {
                Uco = Convert.ToInt32(lineParts[0]),
                LastName = lineParts[1],
                FirstName = lineParts[2]
            };
            Students.Add(newStudent);

            for (int index = 3; index < ((lineParts.Length - 3) / 2) + 3; index += 2)
            {
                string subjectCode = lineParts[index];
                string markString = lineParts[index + 1];

                Seminar seminar = seminars.SingleOrDefault(x => x.Code == subjectCode);

                if (seminar != null)
                {
                    if (Regex.IsMatch(markString, "A|B|C|D|E|F", RegexOptions.IgnoreCase))
                    {
                        Mark mark;
                        Enum.TryParse<Mark>(lineParts[index + 1], out mark);
                        newStudent.Marks.Add(seminar, mark);
                    }
                    else if (Regex.IsMatch(markString, @"\?"))
                    {
                        newStudent.Marks.Add(seminar, null);
                    }
                }
            }
        }

        public void SaveToFile(string filename)
        {
            FileStream dataFile = null;
            StreamWriter writer = null;
            try
            {
                dataFile = new FileStream(filename, FileMode.Create, FileAccess.Write);
                writer = new StreamWriter(dataFile);
                foreach (var student in students)
                {
                    writer.WriteLine(GenerateLine(student));
                }
            }
            finally
            {
                if (writer != null)
                    writer.Close();
                if (dataFile != null)
                    dataFile.Close();
            }
        }

        private static string GenerateLine(Student student)
        {
            StringBuilder b = new StringBuilder();

            b.Append(student.Uco);
            b.Append(FileValuesSeparator);
            b.Append(student.LastName);
            b.Append(FileValuesSeparator);
            b.Append(student.FirstName);

            foreach (var mark in student.Marks)
            {
                b.Append(FileValuesSeparator);
                b.Append(mark.Key.Code);
                b.Append(FileValuesSeparator);
                b.Append(mark.Value.Value);
            }
            return b.ToString();
        }
    }
}
