﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;

namespace PV178.Tutorial3.SpringfieldUniversitySimple
{
    public class University
    {
        private const char FileValuesSeparator = ',';
        private const double BestStudentsAverageLimit = 1.2;
        private const double BadStudentAverageLimit = 3;

        private readonly IList<Student> students = new List<Student>();

        public IList<Student> Students { get { return students; } }

        /*
         * Generator
         * 
         * Pomoci klicoveho slova yield lze vytvorit metodu, ktera se chova jako 
         * iterator. Kompilator sam vygeneruje prislusnou tridu pro popsany iterator.
         * 
         * yield return <vyraz>;   Oznacuje vraceni hodnoty z iteratoru a zaroven
         *                         si zapamatuje pozici v programu iteratoru, kde
         *                         skoncil. V dalsi iteraci pokracuje od tohoto
         *                         mista dale.
         *                           
         * yield break;            Slouzi pro okamzite ukonceni iterace. Na rozdil 
         *                         samotneho break se okamzite ukonci cely iterator,
         *                         ne jen prislusny cyklus.
         * 
         */

        public IEnumerable<Student> GetBestStudents()
        {
            foreach (var student in students.Where(s => s.Marks.Any(x => x.Value.HasValue)))
            {
                if (student.Marks.Where(m => m.Value.HasValue).Average(x => x.Value.ToScore()) <= BestStudentsAverageLimit)
                {
                    yield return student;
                }
            }
        }

        public IEnumerable<Student> GetWorstStudents()
        {
            foreach (var student in this.students.Where(s => s.Marks.Any(x => x.Value.HasValue)))
            {
                if (student.Marks.Where(m => m.Value.HasValue).Average(x => x.Value.ToScore()) >= BadStudentAverageLimit)
                {
                    yield return student;
                }
            }
        }

        /*
         * LINQ
         */

        public IEnumerable<Student> StudentsWithMissingMarks()
        {
            return from s in students
                   where s.Marks.Any(x => x.Value.HasValue == false)
                   select s;
        }

        public void PrintStudentsGroupedByLastName()
        {
            var query = students
                .GroupBy(s => s.LastName[0])
                .OrderBy(g => g.Key)
                .Select(g => new
                {
                    FirstCharacterOfLastName = g.Key,
                    Students = g
                });

            foreach (var item in query)
            {
                Console.WriteLine(item.FirstCharacterOfLastName);
                foreach (var student in item.Students)
                {
                    Console.WriteLine("- {0}, {1} ({2})", student.LastName, student.FirstName, student.Uco);
                }
            }
        }

        public void PrintMarksGroupedBySeminars()
        {
            var query = from s in students
                        from m in s.Marks
                        where m.Value.HasValue
                        group m by m.Key into marksBySeminar
                        select new { Seminar = marksBySeminar.Key, Marks = marksBySeminar };

            foreach (var item in query)
            {
                Console.WriteLine(item.Seminar);
                Console.WriteLine("- Average: {0:F1}", item.Marks.Average(m => m.Value.ToScore()));
                Console.WriteLine("- Best:    {0}", item.Marks.Min(m => m.Value));
                Console.WriteLine("- Worst:   {0}", item.Marks.Max(m => m.Value));
            }
        }

        /*
         * File Operations
         *
         */

        public void LoadFromFile(string filename)
        {
            // kontrola parametru a try-finally pro prehlednost vynechany

            FileStream dataFile = new FileStream(filename, FileMode.Open, FileAccess.Read);
            StreamReader reader = new StreamReader(dataFile);

            string line = reader.ReadLine();
            while (line != null)
            {
                ParseLine(line);
                line = reader.ReadLine();
            }

            reader.Close();
            dataFile.Close();
        }

        private void ParseLine(string line)
        {
            string[] values = line.Split(FileValuesSeparator);

            Student newStudent = new Student
            {
                Uco = Convert.ToInt32(values[0]),
                LastName = values[1],
                FirstName = values[2]
            };
            Students.Add(newStudent);

            for (int index = 3; index < values.Length; index += 2)
            {
                string subjectCode = values[index];
                string markString = values[index + 1];

                if (Regex.IsMatch(markString, "A|B|C|D|E|F", RegexOptions.IgnoreCase))
                {
                    Mark mark;
                    Enum.TryParse<Mark>(values[index + 1], out mark);
                    newStudent.Marks.Add(subjectCode, mark);
                }
                else if (Regex.IsMatch(markString, @"\?"))
                {
                    newStudent.Marks.Add(subjectCode, null);
                }
            }
        }

        public void SaveToFile(string filename)
        {
            // kontrola parametru a try-finally pro prehlednost vynechany

            FileStream dataFile = new FileStream(filename, FileMode.Create, FileAccess.Write);
            StreamWriter writer = new StreamWriter(dataFile);

            foreach (var student in students)
            {
                writer.WriteLine(GenerateLine(student));
            }

            writer.Close();
            dataFile.Close();
        }

        private static string GenerateLine(Student student)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(student.Uco);
            builder.Append(FileValuesSeparator);
            builder.Append(student.LastName);
            builder.Append(FileValuesSeparator);
            builder.Append(student.FirstName);

            foreach (var mark in student.Marks)
            {
                builder.Append(FileValuesSeparator);
                builder.Append(mark.Key);
                builder.Append(FileValuesSeparator);
                if (mark.Value.HasValue)
                {
                    builder.Append(mark.Value);
                }
                else
                {
                    builder.Append("?");
                }
            }
            return builder.ToString();
        }
    }
}
